# -*- coding: utf-8 -*-
#!/usr/bin/env python3
"""Convert Valgrind XML to Code Quality JSON

Valgrind is a useful tool detect memory leak in C/C++ code.
Developer tools, such as GitLab, can display useful insights about code quality,
when given a JSON report file defined by Code Quality's GitLab.

This tool convert Valgrind XML report to Code Quality JSON.

Example:
    # Generate valgrind report as XML
    valgrind --tool=memcheck --show-leak-kinds=all --track-origins=yes --verbose --xml=yes --xml-file=valgrind_out.xml your_exe
    # Convert to a Code Climate JSON report
    valgrind-codequality --input-file valgrind_out.xml --output-file valgrind.json

References:
  - https://valgrind.org/
  - https://docs.gitlab.com/ee/user/project/merge_requests/code_quality.html#implementing-a-custom-tool

SPDX-License-Identifier: MIT
"""

import hashlib
import json
import logging
import os
import typing
from copy import deepcopy

# third-party
import xmltodict

__version__ = "1.0.3"

log = logging.getLogger(__name__)

# Source: https://github.com/codeclimate/platform/blob/master/spec/analyzers/SPEC.md#data-types
CODE_QUAL_ELEMENT = {
    "severity": "",
    "description": "",
    "fingerprint": "",
    "location": {"path": "", "lines": {"begin": -1}},
}


def _get_codeclimate_severity(valgrind_severity: str) -> str:
    """Get Code Climate severity, from valgrind severity string

    CodeQuality: info, minor, major, critical, blocker
    """
    severity = ""
    if "Leak_DefinitelyLost" == valgrind_severity:
        severity = "blocker"
    else:
        if valgrind_severity in [
            "UninitCondition",
            "UninitValue",
            "ClientCheck",
            "Overlap",
            "Leak_StillReachable",
        ]:
            severity = "minor"
        elif valgrind_severity in [
            "SyscallParam",
            "InvalidFree",
            "MismatchedFree",
            "InvalidJump",
            "InvalidMemPool",
            "Leak_PossiblyLost",
        ]:
            severity = "major"
        elif valgrind_severity in [
            "InvalidRead",
            "InvalidWrite",
            "Leak_IndirectlyLost",
        ]:
            severity = "critical"
        else:
            severity = "info"

    return severity


def convert_file(
    fname_in: str,
    fname_out: str,
    file_concat: typing.Optional[str] = None,
    base_dirs: typing.Optional[list] = None,
    relative_dir: typing.Optional[str] = None,
) -> int:
    """Convert valgrind XML file to GitLab-compatible "Code Quality" JSON report

    Args:
        fname_in (str):
          Input file path (valgrind XML). Like 'valgrind.xml'.
        fname_out (str):
          Output file path (code quality JSON). Like 'valgrind.json'.
        file_concat (str):
          File path (JSON) contains code quality informations from previous analyze.
        base_dirs (list):
          Base directories where source files with relative paths can be found.
        relative_dir (str):
          Absolute path to change absolute path to relative from this path.

    Returns:
        int: If processing failed, a negative value. If successful, number of
          valgrind issues processed.
    """
    fin = None
    dict_concat = None
    json_out_str = ""
    num_cq_issues_converted = 0

    # Read concat file
    if file_concat:
        file_concat = os.path.abspath(file_concat)
        if not os.path.isfile(file_concat):
            log.error(
                "Concat file (JSON) file does not exist or cannot be opened -- '%s'",
                file_concat,
            )
            return -1
        with open(file_concat, mode="r", encoding="utf-8") as fin_concat:
            dict_concat = json.loads(fin_concat.read())

    # Convert Input
    fname_in = os.path.abspath(fname_in)
    if not os.path.isfile(fname_in):
        log.error(
            "Input (valgrind XML) file does not exist or cannot be opened -- '%s'",
            fname_in,
        )
        return -1

    log.debug("Reading input file: %s", fname_in)
    with open(fname_in, mode="rt", encoding="utf-8", errors="backslashreplace") as fin:
        json_out_str, num_cq_issues_converted = _convert(
            fin.read(),
            base_dirs=base_dirs,
            relative_dir=relative_dir,
            dict_concat=dict_concat,
        )

    log.debug("Writing output file: %s", fname_out)
    with open(fname_out, "w", encoding="utf-8") as f_out:
        f_out.write(json_out_str)

    return num_cq_issues_converted


def _get_line_from_file(filename: str, line_number: int) -> str:
    """Return a specific line in a file as a string.

    I've found that linecache.getline() will end up raising a UnicodeDecodeError
    if the source file we're opening has non-UTF-8 characters in it. So, here,
    we're explicitly escaping those bad characters.

    Side note, it seems valgrind v2.0+ will generate a 'syntaxError' for
    "unhandled characters", so you could find these issues with your source code
    more easily.

    Args:
        filename (str):
          Name of file to open and read line from
        line_number (int):
          Number of the line to extract. Line number starts at 1.

    Returns:
        str: Contents of the specified line.
    """
    max_line_cnt = 0
    if line_number <= 0:
        return str(filename) + "<the whole file>"

    filename = os.path.abspath(filename)
    if not os.path.isfile(filename):
        raise FileNotFoundError(
            "Source code file does not exist or cannot be opened. Missing a base directory?\n--> '{}'".format(
                filename
            )
        )

    with open(filename, mode="rt", encoding="utf-8", errors="backslashreplace") as fin:
        for i, line in enumerate(fin):
            if (i + 1) == line_number:
                # log.debug("Extracted line %s:%d", filename, line_number)
                return line
            max_line_cnt += 1

    log.warning(
        "Only %d lines in file. Can't read line %d from '%s'",
        max_line_cnt,
        line_number,
        filename,
    )
    return "Can't read line {} from a {} line file".format(line_number, max_line_cnt)


def _convert(
    xml_input,
    base_dirs: typing.Optional[typing.List[str]] = None,
    relative_dir: typing.Optional[str] = None,
    dict_concat: typing.Optional[typing.Dict[str, str]] = None,
) -> typing.Tuple[str, int]:
    """Convert valgrind XML to Code Climate JSON

    Note:
        There isn't a great 1:1 conversion from valgrind's "severity" level, to
        the Code Climate's "categories." To prevent information loss, the
        original valgrind severity is appended to the category list.

        In the future, maybe this conversion can be made using valgrind's "id"
        or check name.

    Args:
        xml_input (str): Filename of the XML from valgrind
        base_dirs (list):
          Base directories where source files with relative paths can be found.
        relative_dir (str):
          Absolute path to change absolute path to relative from this path.
        dict_concat: dictionnary contains code quality informations from previous analyze.

    Returns:
        Tuple, where the first element, a string, is the JSON conversion result
        and the second element, an int, is the number of issues converted.
    """

    dict_in = xmltodict.parse(xml_input=xml_input)

    if len(dict_in) == 0:
        log.info("Empty file imported. Skipping...")
        return ("[ ]", 0)

    dict_out = []
    fingerprints = []

    # Add pevious analyze
    if dict_concat:
        dict_out = dict_concat

    # Ensure this XML report has errors to convert
    if not "error" in dict_in["valgrindoutput"]:
        log.warning("No <error> in XML file. Nothing to do.")
        return (json.dumps(dict_out), 0)

    errors = []
    if not isinstance(dict_in["valgrindoutput"]["error"], list):
        errors = [dict_in["valgrindoutput"]["error"]]
    else:
        errors = dict_in["valgrindoutput"]["error"]

    for error in errors:

        log.debug("Processing -- %s", str(error))

        # Some information messages are not related to the code.
        # Let's let the user know, then skip.
        if "stack" not in error:
            log.info("No stack. Skipping the below issue:\n  %s", error["unique"])
            continue

        # Extract frames
        stacks = error["stack"]
        if not type(stacks) is list:
            stacks = [stacks]

        tmp_dict = dict(CODE_QUAL_ELEMENT)
        kind = error["kind"]
        if "xwhat" in error:
            tmp_dict["description"] = error["xwhat"]["text"]
        else:
            tmp_dict["description"] = error["what"]

        tmp_dict["description"] = kind + ": " + tmp_dict["description"]
        tmp_dict["severity"] = _get_codeclimate_severity(kind)

        path = ""
        line = -1
        for index in range(len(stacks[0]["frame"])):
            frame = stacks[0]["frame"][index]
            first_frames = stacks[0]["frame"][index]
            if "dir" in frame and "file" in frame:
                path = os.path.join(frame["dir"], frame["file"])
                line = int(first_frames["line"])

                # File path not found
                if not os.path.isfile(path):
                    path = ""
                    # Check in base dirs
                    if isinstance(base_dirs, list):
                        for d in base_dirs:
                            tmp_path = os.path.join(d, frame["file"])
                            if os.path.isfile(tmp_path):
                                path = tmp_path
                                break

                if path == "":
                    continue
                else:
                    break

        # Path found
        if path != "":
            if relative_dir:
                tmp_dict["location"]["path"] = os.path.relpath(path, relative_dir)
            else:
                tmp_dict["location"]["path"] = path
            tmp_dict["location"]["lines"]["begin"] = line

            # GitLab requires the fingerprint field. Code Climate describes this as
            # being used to uniquely identify the issue, so users could "exclude it
            # from future analysis."
            #
            # The components of the fingerprint aren't well defined, but Code Climate
            # has some examples here:
            # https://github.com/codeclimate/codeclimate-duplication/blob/1c118a13b28752e82683b40d610e5b1ee8c41471/lib/cc/engine/analyzers/violation.rb#L83
            # https://github.com/codeclimate/codeclimate-phpmd/blob/7d0aa6c652a2cbab23108552d3623e69f2a30282/tests/FingerprintTest.php

            codeline = _get_line_from_file(filename=path, line_number=line).strip()

            fingerprint_str = "valgrind-" + kind + "-" + path + "-" + codeline
            log.debug("Fingerprint string: '%s'", fingerprint_str)
            tmp_dict["fingerprint"] = hashlib.md5(
                (fingerprint_str).encode("utf-8")
            ).hexdigest()

            # Append this record if not already exists
            if not fingerprint_str in fingerprints:
                dict_out.append(deepcopy(tmp_dict))
                fingerprints.append(fingerprint_str)

    if len(dict_out) == 0:
        log.warning("Result is empty")
    return (json.dumps(dict_out, indent=4), len(dict_out))


if __name__ == "__main__":
    import warnings

    warnings.warn(
        "use 'python3 -m valgrind_codequality', not 'python3 -m valgrind_codequality.__init__'",
        DeprecationWarning,
    )



# -*- coding: utf-8 -*-
#!/usr/bin/env python3

import argparse
import logging
import sys
import os

from . import __version__, convert_file


def _init_logging(level, fname):
    """Setup root logger to log to console, when this is run as a script"""
    logging.getLogger().setLevel(level)

    h_console = logging.StreamHandler()
    log_fmt_short = logging.Formatter(
        "[%(relativeCreated)d] %(name)s %(levelname)s: %(message)s", datefmt="%H:%M:%S"
    )
    h_console.setFormatter(log_fmt_short)

    # Add console handler to root logger
    logging.getLogger("").addHandler(h_console)

    if fname != "/dev/null":
        h_file = logging.FileHandler(fname, encoding="utf-8")
        log_fmt_long = logging.Formatter(
            "%(asctime)s %(name)s %(levelname)s: %(message)s"
        )
        h_file.setFormatter(log_fmt_long)
        # Add file handler to root logger
        logging.getLogger("").addHandler(h_file)


def _get_args() -> argparse.Namespace:
    """Parse CLI args with argparse"""
    # Make parser object
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "-i",
        "--input-file",
        metavar="INPUT_XML_FILE",
        dest="input_file",
        type=str,
        default="valgrind.xml",
        help="the valgrind XML output file to read defects from (default: %(default)s)",
    )

    parser.add_argument(
        "-o",
        "--output-file",
        metavar="FILE",
        dest="output_file",
        type=str,
        default="valgrind.json",
        help="output filename to write JSON to (default: %(default)s)",
    )

    parser.add_argument(
        "-c",
        "--concatenate",
        metavar="FILE_CONCAT",
        type=str,
        default="",
        help="concatenate output with this code quality file",
    )

    parser.add_argument(
        "-l",
        "--loglevel",
        metavar="LVL",
        type=str,
        choices=["debug", "info", "warn", "error"],
        default="info",
        help="set logging message severity level (default: '%(default)s')",
    )

    parser.add_argument(
        "-L",
        "--logfile",
        metavar="FNAME",
        type=str,
        default="/dev/null",
        help="log messages to a file, in addition to the console (default: '%(default)s')",
    )

    parser.add_argument(
        "-v",
        "--version",
        dest="print_version",
        action="store_true",
        help="print version and exit",
    )

    parser.add_argument(
        "-b",
        "--base-dirs",
        metavar="base_dirs",
        action="append",
        default=[],
        help="base directories where source code files can be found. (default: '%(default)s')",
    )

    parser.add_argument(
        "-r",
        "--relative-dir",
        metavar="RELATIVE_DIR",
        type=str,
        default=os.getcwd(),
        help="change absolute path to relative from this path. (default: '%(default)s')",
    )

    return parser.parse_args()


def main() -> int:
    """Convert a valgrind XML file to Code Climate JSON file, at the command line."""

    if sys.version_info < (3, 6, 0):
        sys.stderr.write("You need python 3.6 or later to run this script\n")
        return 1

    args = _get_args()

    _init_logging(level=args.loglevel.upper(), fname=args.logfile)
    m_log = logging.getLogger(__name__)

    if args.print_version:
        print(__version__)
        return 0

    ret = convert_file(
        fname_in=args.input_file,
        fname_out=args.output_file,
        file_concat=args.concatenate,
        base_dirs=["."] if len(args.base_dirs) == 0 else args.base_dirs,
        relative_dir=args.relative_dir,
    )
    if ret < 0:
        m_log.error("Conversion failed")
        return 1

    # Logging the total count.
    # Side note: In personal projects, I intent to `cat` the log file and extract
    # this number, to record in a "metrics.txt" file and to generate a badge with "anybadge"
    # See:
    #   - https://docs.gitlab.com/ee/ci/metrics_reports.html
    #   - https://pypi.org/project/anybadge
    m_log.info("Converted %d valgrind issues", ret)

    return 0


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['valgrind_codequality']

package_data = \
{'': ['*']}

install_requires = \
['xmltodict>=0.13.0,<0.14.0']

entry_points = \
{'console_scripts': ['valgrind-codequality = '
                     'valgrind_codequality.__main__:main']}

setup_kwargs = {
    'name': 'valgrind-codequality',
    'version': '1.0.3',
    'description': 'Convert Valgrind XML to GitLab Code Quality JSON file.',
    'long_description': '# valgrind-codequality\n\n[![badge-pypi](https://img.shields.io/pypi/v/valgrind-codequality.svg?logo=pypi)](https://pypi.python.org/pypi/valgrind-codequality/)\n&nbsp;\n[![badge-pypi-downloads](https://img.shields.io/pypi/dm/valgrind-codequality)](https://pypi.org/project/valgrind-codequality/)\n\n\n[![badge-pipeline](https://gitlab.com/echopouet/valgrind-codequality/badges/main/pipeline.svg)](https://gitlab.com/echopouet/valgrind-codequality/-/pipelines?scope=branches)\n&nbsp;\n[![badge-coverage](https://gitlab.com/echopouet/valgrind-codequality/badges/main/coverage.svg)](https://gitlab.com/echopouet/valgrind-codequality/-/pipelines?scope=branches)\n&nbsp;\n[![badge-pylint](https://gitlab.com/echopouet/valgrind-codequality/-/jobs/artifacts/main/raw/badge.svg?job=pylint)](https://gitlab.com/echopouet/valgrind-codequality/-/pipelines?scope=branches)\n&nbsp;\n[![badge-formatting](https://gitlab.com/echopouet/valgrind-codequality/-/jobs/artifacts/main/raw/badge.svg?job=format_black)](https://gitlab.com/echopouet/valgrind-codequality/-/pipelines?scope=branches)\n&nbsp;\n[![badge-issues-cnt](https://img.shields.io/badge/dynamic/json?label=issues&query=statistics.counts.opened&url=https%3A%2F%2Fgitlab.com%2Fapi%2Fv4%2Fprojects%2F19114200%2Fissues_statistics%3Fscope%3Dall)](https://gitlab.com/echopouet/valgrind-codequality/-/issues)\n\n\n## About\n\nI wanted reports from [Valgrind](https://valgrind.org/) to appear in GitLab Merge Requests as [Code Quality reports](https://docs.gitlab.com/ee/user/project/merge_requests/code_quality.html#implementing-a-custom-tool), which is a JSON file defined by the Code Quality\'s GitLab.\n\nThat\'s all this does: convert Valgrind XML report to Code Quality JSON.\n\nContributions are welcome.\n\n[!["Buy Me A Coffee"](https://www.buymeacoffee.com/assets/img/custom_images/yellow_img.png)](https://www.buymeacoffee.com/EchoPouet)\n\n### Usage\n\nIt is primarily used as a console script. As such, ensure you have Python 3\'s "scripts" directory in your `PATH` variable.\nFor example, on Linux, that might be `$HOME/.local/bin`.\n\nTo test, try the `--help` or `--version` flags:\n```bash\nvalgrind-codequality --help\n```\n\nThis script follows that example and provides similar command-line options.\nA typical workflow might look like this:\n\n```bash\n# Generate valgrind report as XML\nvalgrind --tool=memcheck --leak-check=full --show-leak-kinds=all --track-origins=yes --verbose --xml=yes --xml-file=valgrind_out.xml your_exe\n# Convert to a Code Climate JSON report\nvalgrind-codequality --input-file valgrind_out.xml --output-file valgrind.json\n```\n\nIf you wanted, you could invoke the script directly as a module, like this:\n\n```bash\n# Run as a module instead (note the underscore in the module name here)\npython -m valgrind_codequality --input-file=valgrind_out.xml --output-file=valgrind.json\n```\n\nNow, in your GitLab CI script, [upload this file](https://docs.gitlab.com/ee/ci/pipelines/job_artifacts.html#artifactsreportscodequality)\nas a Code Quality report.\n\n```yaml\nmy-code-quality:\n  script:\n    - [...]\n  artifacts:\n    reports:\n      codequality: valgrind.json\n```\n\n### Contributing\n\n* Sign the contributor agreement (coming soon)\n* Format with [black](https://pypi.org/project/black/)\n* Check with [pylint](https://pypi.org/project/pylint/)\n\n### Credits & Trademarks\n\nvalgrind is an open-source project with a GPL v3.0 license.\n* http://valgrind.sourceforge.net\n* https://github.com/danmar/valgrind\n\n"GitLab" is a trademark of GitLab B.V.\n* https://gitlab.com\n* https://docs.gitlab.com/ee/user/project/merge_requests/code_quality.html\n\nAll other trademarks belong to their respective owners.\n',
    'author': 'Arnaud Moura',
    'author_email': 'arnaudmoura@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.8,<=3.12',
}


setup(**setup_kwargs)
